# File generated from our OpenAPI spec
from __future__ import absolute_import, division, print_function

from ten99policy import api_resources


OBJECT_CLASSES = {
    # data structures
    api_resources.ListObject.OBJECT_NAME: api_resources.ListObject,
    api_resources.Contractors.OBJECT_NAME: api_resources.Contractors,
    api_resources.Entities.OBJECT_NAME: api_resources.Entities,
    api_resources.Jobs.OBJECT_NAME: api_resources.Jobs,
    api_resources.Policies.OBJECT_NAME: api_resources.Policies,
    api_resources.Quotes.OBJECT_NAME: api_resources.Quotes,
}
